;.include "inc/tn2313def.inc"
.device ATtiny2313

;FUSE_H=0x9f
;FUSE_L=0xcf

;Pins, port d
.equ rxd	=	0
.equ txd	=	1
.equ photo	=	2
.equ laser	=	5


;Pins, port b
.equ motor	=	2

;Used registers
.def dat	=	r0
.def maxl	=	r2
.def maxh	=	r3
.def temp	=	r16
.def temp2	=	r17
.def temp3	=	r18
.def posx	=	r19
.def posy	=	r20
.def wval	=	r21
.def flags	=	r22
.def ylo	=	r28
.def zlo	=	r30
.def zhi	=	r31

;Flags is a bitfield, with the following bits defined:
.equ f_ignt1ovf	=	0
.equ f_rectimings=	1
.equ f_overtime	=	2
.equ f_desync	=	3

.equ NOPIXELS=96 ;Number of pixels displayed.

.org 0
;Interrupt-table
    rjmp reset		;Reset
    rjmp gotbeam	;Int0
    reti		;Int1
    reti		;timer 1 capture
    reti		;timer 1 compare A
    rjmp error		;timer 1 ovf
    reti		;timer 0 ovf
    reti		;uart rx complete
    reti		;uart data register empty
    reti		;uart tx complete
    reti		;analog comparator
    reti		;pin change int
    reti		;timer 1 compare B
    reti		;timer 0 compare A
    reti		;timer 0 compare B
    reti		;USI start
    reti		;USI overflow
    reti		;EEPROM ready
    reti		;Watchdog ovf

    
reset:
    ldi temp,RAMEND
    out spl,temp

    ;Initialize i/o-pins
    ldi temp,0b0100010
    out ddrd,temp
    ldi temp,0b000100
    out ddrb,temp

    ;disable ocr0a<->timer-connection
    ldi temp,0
    out tccr0a,temp

    ;Let motor spin up
    sbi portb,motor
    ldi temp,0
waitmotor1:
    ldi temp2,0
waitmotor2:
    ldi temp3,80
waitmotor3:
    dec temp3
    brne waitmotor3
    dec temp2
    brne waitmotor2
    dec temp
    brne waitmotor1

    ;Init timer 0 as a pwm channel for the motor speed
    ldi temp,0b10000011
    out tccr0a,temp
    ldi temp,3
    out tccr0b,temp
    ldi temp,10
    out eear,temp
    sbi eecr,0
    in temp,eedr
    out ocr0a,temp
    
    

    ;Init timer/counter 1 to count at clk/64
    ldi temp,0
    out tccr1a,temp
    ldi temp,0b11
    out tccr1b,temp
    ldi temp,(1<<7)
    out timsk,temp
    
    ;Init int 0 to trigger on low-to-high transition, wich happens as soon
    ;as the laser has hit and left the phototransistor
    ldi temp,0b11
    out mcucr,temp
    ldi temp,(1<<6)
    out gimsk,temp
    
    
    
    ;Init uart
    ldi temp,0
    out ucsra,temp
    ldi temp,0b011000
    out ucsrb,temp
    ldi temp,0b0110
    out ucsrc,temp
    ;9600 baud @ 12.288MHz xtal.
    ldi temp,0
    out ubrrh,temp
    ldi temp,79
    out ubrrl,temp
    

    ;Write init msg
    ldi zlo,low(strStart*2)
    ldi zhi,high(strStart*2)
    ldi ylo,RAMEND-127
    ldi temp3,NOPIXELS/8
writeinitstr:
    lpm
    adiw zlo,1
    mov temp,dat
    st y+,temp
    dec temp3
    brne writeinitstr
    
    ;Reset positional variables & enable display subsystem
    ldi posx,0
    ldi posy,0
    ldi ylo,0
    sbi portd,laser    
    sei

;---------  Main loop  ------------------
;The menu and serial routines live here
mainloop:
    rcall recvbyte
    rcall sendbyte
    cpi temp,27 ;escape
    brne ndomenu
    rjmp domenu
ndomenu:
    cpi temp,13
    breq IsEnter
    cpi temp,10
    breq IsEnter
    ;filter undisplayable chars
    cpi temp,32
    brlo mainloop
    cpi temp,129
    brsh mainloop
    cpi ylo,0
    brne NoClear
    
    ldi ylo,RAMEND-127
    ldi temp2,32
DoClear:
    st y+,temp2
    cpi ylo,RAMEND-32
    brne DoClear
    
    ldi ylo,RAMEND-127
NoClear:
    st y+,temp
    cpi ylo,RAMEND-32
    brne mainloop
    dec ylo
    rjmp mainloop

IsEnter:
    ldi ylo,0
    rjmp mainloop

;This menu sub is moved higher to make everything stay in brne reach. The
;main menu function is underneath this.

;Adjust motor speed (by adjusting ocr0a)
domotoradj:
    cli
    ldi zlo,low(strMadj*2)
    ldi zhi,high(strMadj*2)
    rcall sendstr
madjloop:
    rcall recvbyte
    in temp2,ocr0a
    inc temp2
    cpi temp,']'
    breq domadj
    dec temp2
    dec temp2
    cpi temp,'['
    breq domadj
    cpi temp,27
    breq madjfinished
    rjmp madjloop
domadj:
    mov temp,temp2
    rcall sendhex
    ldi temp,13
    rcall sendbyte
    ldi temp,10
    rcall sendbyte
    out ocr0a,temp2
    rjmp madjloop
    
madjfinished:
    ;save to eeprom
    in temp2,ocr0a
    ldi temp,10
    out eear,temp
    out eedr,temp2
    ldi temp,(1<<2)
    out eecr,temp
    sbi eecr,1
madjw:
    sbic eecr,1
     rjmp madjw

    sbr flags,(1<<f_ignt1ovf)
    sei
    ;wait for the int to happen
    nop
    nop
    nop
    nop
    cbr flags,(1<<f_ignt1ovf)
    rjmp domenu


;Show menu    
domenu:
    ldi zlo,low(strMenu*2)
    ldi zhi,high(strMenu*2)
    rcall sendstr
    rcall recvbyte
    ;Calibrate channel?
    cpi temp,'1'
    brlo nocal
    cpi temp,'8'
    brsh nocal
    rjmp docalibrate    
nocal:
    cpi temp,'8'
    breq dolaseron
    cpi temp,'9'
    breq doshowtimings
    cpi temp,'0'
    breq doshowstate
    cpi temp,'m'
    breq domotoradj
    cpi temp,'M'
    breq domotoradj
    cpi temp,27
    brne domenu
    ldi zlo,low(strCalDone*2)
    ldi zhi,high(strCalDone*2)
    rcall sendstr    
    rjmp mainloop


;Menu option 8: laser continuously on
dolaseron:
    cli
    sbi portd,laser
    ldi zlo,low(strLon*2)
    ldi zhi,high(strLon*2)
    rcall sendstr
    rcall recvbyte
    ldi temp3,0
    out tcnt1h,temp3
    out tcnt1l,temp3
    sbr flags,(1<<f_ignt1ovf)
    sei
    ;wait for the int to happen
    nop
    nop
    nop
    nop
    cbr flags,(1<<f_ignt1ovf)
    rjmp domenu

doshowstate:
    ldi temp,'.'
    sbrc flags,f_desync
     ldi temp,'d'
    sbrc flags,f_overtime
     ldi temp,'t'
    rcall sendbyte

    sbis ucsra,7
     rjmp doshowstate
    in temp,udr
    cbr flags,(1<<f_rectimings)
    rjmp domenu


;Menu option 9: show timings
doshowtimings:
    sbr flags,(1<<f_rectimings)
    
    ldi zlo,RAMEND-80;
    ldi temp2,7
printtimingloop:
    ld temp,z+
    rcall sendhex
    ld temp,z+
    rcall sendhex
    ldi temp,' '
    rcall sendbyte
    dec temp2
    brne printtimingloop

    ldi temp,13
    rcall sendbyte
    ldi temp,10
    rcall sendbyte
    
    sbis ucsra,7
     rjmp doshowtimings
    in temp,udr
    cbr flags,(1<<f_rectimings)
    rjmp domenu
    


docalibrate:
;Fill up half the display field with '|' characters
    ldi ylo,RAMEND-127
    ldi temp2,'|'
DoClearCal:
    st y+,temp2
    cpi ylo,RAMEND-32
    brne DoClearCal
    ldi ylo,0

;and the other half with spaces
    ldi ylo,RAMEND-127+(NOPIXELS/16)
    ldi temp2,' '
DoClearCal2:
    st y+,temp2
    cpi ylo,RAMEND-32
    brne DoClearCal2
    ldi ylo,0

;Show the 'calibrating channel suchandsuch' message
    ldi zlo,low(strCal*2)
    ldi zhi,high(strCal*2)
    rcall sendstr
    rcall sendbyte
    ldi zlo,low(strCal2*2)
    ldi zhi,high(strCal2*2)
    rcall sendstr
    
    mov temp3,temp
    subi temp3,'1' ;now temp3 contains the channel number in binary
calloop:
    cli ;needed because the display routine interrupt accesses the eeprom too
    ;load current cal val
    out eear,temp3
    sbi eecr,0
    in temp,eedr
    sei
    com temp
    rcall sendhex
    ;next line
    ldi temp,10
    rcall sendbyte
    ldi temp,13
    rcall sendbyte
    ;get byte and process
    rcall recvbyte
    cpi temp,'['
    breq caldomore
    cpi temp,']'
    breq caldoless
    cpi temp,27
    breq calend
    rjmp calloop
caldomore:
    ;read cal val
    cli
    out eear,temp3
    sbi eecr,0
    in temp,eedr
    sei
    com temp ;cal vals are stored as 1s complement
    inc temp ;increase
    ;limit
    cpi temp,150
    brlo calnottoomuch
    ldi temp,150
calnottoomuch:
    com temp ;1-complement again
    ;and write
    cli
    out eear,temp3
    out eedr,temp
    ldi temp,(1<<2)
    out eecr,temp
    sbi eecr,1
calmorew:
    sbic eecr,1
     rjmp calmorew
    sei
    ;Done
    rjmp calloop

;Same as caldomore
caldoless:
    cli
    out eear,temp3
    sbi eecr,0
    in temp,eedr
    sei
    com temp
    dec temp
    cpi temp,$ff
    brne calnottooless
    ldi temp,0
calnottooless:
    com temp
    cli
    out eear,temp3
    out eedr,temp
    sbi eecr,2
    sbi eecr,1
callessw:
    sbic eecr,1
     rjmp callessw
    sei
    rjmp calloop
    
calend:
    ldi zlo,low(strCalDone*2)
    ldi zhi,high(strCalDone*2)
    rcall sendstr
    rjmp domenu

;----------------------  String data -------------
strMenu:
    .db 13,10,"Menu:",13,10,"1-7: Calibrate channel",13,10,"8: laser always on",13,10,"9: show timings",13,10,"0: show state",13,10,"m: motor speed adjust",13,10,"esc: exit menu",13,10,0
strCal:
    .db "Calibrating ",0
strCal2:
    .db ", use [, ], esc to save.",13,10,0
strCalDone:
    .db "OK",13,10,0
strLon:
    .db "Any key to resume...",13,10,0
strMadj:
    .db "[ and ] adjust motor speed; esc saves.",13,10,0
strError:
    .db 13,10,"Motor error! Esc reboots, I ignores",13,10,0


strStart: 
    .db "LaserDisplay"
    
;------------------- Utility subroutines  -------------
;Sends the zero-terminated string pointed at by Z to the serial port
sendstr:
    push temp
nextchar:
    lpm
    mov temp,dat
    cpi temp,0
    breq stringend
    rcall sendbyte
    adiw zlo,1
    rjmp nextchar
stringend:
    pop temp
    ret

;Sends the value in temp as hex to the serial port
sendhex:
    push temp
    swap temp
    rcall sendnibble
    pop temp
    rcall sendnibble
    ret
    
sendnibble:
    push temp
    andi temp,$0f
    subi temp,-'0'
    cpi temp,'9'+1
    brlo sendnibbleok
    subi temp,-(-'0'+'A'-10)
sendnibbleok:
    rcall sendbyte
    pop temp
    ret
    
;Receive a byte
recvbyte:
    sbis ucsra,7
     rjmp recvbyte
    in temp,udr
    ret
    
;Send a byte
sendbyte:
    sbis ucsra,5
     rjmp sendbyte
    out udr,temp
    ret

;----------------   Interrupt routines   ------------------

;This interrupt gets called when the phototransistor gets the beam. It handles
;the beam synchronization and bitmap output.
gotbeam:
    push temp
    in temp,sreg
    push temp
    push temp2
    push temp3
    push zlo
    push zhi
restartint:

    ;Turn off laser
    cbi portd,laser
    
    ;Get how long this cycle took
    in temp,tcnt1l
    in temp2,tcnt1h
    ;re-start measurement of cycle
    ldi temp3,0
    out tcnt1h,temp3
    out tcnt1l,temp3
    
;Save timings for troubleshooting
    sbrs flags,f_rectimings
     rjmp norectimings
    ldi zlo,RAMEND-80
    add zlo,posy
    add zlo,posy
    st z+,temp2
    st z,temp
norectimings:
    
    ;Higher or lower than the current max value?
    cp maxh,temp2
    brlo currislarger
    brne currissmaller
    cp maxl,temp
    brlo currislarger
    
currissmaller:
    ;Is smaller. Is that supposed to be the way?
    cpi posy,0
    brne compend
    ;This should be period 7->0, but it didn't took the longest of all cycles.
    sbr flags,(1<<f_desync)
    inc posy ;perhaps we're right now...
    ;clear max
    clr maxl
    clr maxh
    ;use this rounds time for an indication for wval
    mov wval,temp
    lsr temp2
    ror wval
    lsr temp2
    ror wval
    lsr temp2
    ror wval
    lsr temp2
    ror wval

    rjmp compend
    
currislarger:
    cpi posy,0 ;period 7->0 and largest?
    breq posisok ;Yes -> that's the way it's supposed to be.
    ;Nope -> just adjust maxl and maxh
    mov maxl,temp
    mov maxh,temp2
    rjmp compend
posisok:
    cbr flags,(1<<f_desync)
    ;Use max of normal rounds as wval
    mov wval,maxl
    lsr maxh
    ror wval
    lsr maxh
    ror wval
    lsr maxh
    ror wval
    lsr maxh
    ror wval
    ;clear max
    clr maxl
    clr maxh
compend:

    ;re-start measurement of cycle
    ldi temp,0
    out tcnt1h,temp
    out tcnt1l,temp

    ;Load the offset value saved to eeprom during calibration
    out eear,posy
    sbi eecr,0
    in temp,eedr
    com temp
    ;wait offset value
    inc temp
woffs:
    mov temp2,wval
woffs1:
    dec temp2
    brne woffs1    
    dec temp
    brne woffs

    ;Main blitting routines
    ldi posx,0
bblit:
    lsr temp ;Next pixel
    ;Do we need to load a new byte of image data?
    mov temp2,posx
    andi temp2,7
    brne notnewbyte
    ;Yep.
    mov temp,posx
    ;/8
    lsr temp
    lsr temp
    lsr temp
    ;We need to display the temp'th character from sram.
    ;Load the char
    ldi zlo,RAMEND-127+(NOPIXELS/8)-1
    sub zlo,temp
    ld temp,z
    subi temp,32 ;characters start at 32
    
    ;Load character data
    ldi zlo,low(font*2)
    ldi zhi,high(font*2)
    ;Pick the character
    ldi temp2,0
    lsl temp
    rol temp2
    lsl temp
    rol temp2
    lsl temp
    rol temp2
    add zlo,temp
    adc zhi,temp2
    ;Pick the line from that character
    ldi temp,7
    ldi temp2,0
    sub temp,posy
    add zlo,temp
    adc zhi,temp2
    ;Load line
    lpm
    mov temp,dat
notnewbyte:
    cbi portd,laser    
    sbrc temp,0
     sbi portd,laser    

    rcall wait
    ;If the timing's way off, we could have been starting on the 
    ;next line already. Check that.
    sbis pind,photo
     rjmp timingiswayoff
    ;Timing's ok -> next pixel
    inc posx
    cpi posx,NOPIXELS
    brne bblit

    ;Back porch
    cbi portd,laser    
    rcall wait
    rcall wait
    rcall wait

    cbr flags,(1<<f_overtime)
gonextline:
    
    ;Get ready for the next line
    inc posy
    cpi posy,7
    brne notrestarty
    ldi posy,0
notrestarty:
    ;turn laser on to detect start of next line
    sbi portd,laser    

    pop zhi
    pop zlo
    pop temp3
    pop temp2
    pop temp
    out sreg,temp
    pop temp
    reti

timingiswayoff:    
;Led got triggered in the middle of the text. Abort current line and re-enter
;interrupt routine
    sbr flags,(1<<f_overtime)
    rjmp gonextline

;Subroutine to wait for the beam to advance one pixel
wait:
    push temp

    mov temp,wval
w1:
    dec temp
    brne w1

    pop temp
    ret

;Timer1 overflowed, which means we've gone too long before detecting the laser.
error:
    sbrc flags,f_ignt1ovf
     reti
    cbi portd,laser
    ldi zlo,low(strError*2)
    ldi zhi,high(strError*2)
    rcall sendstr
    
errorhalt:
    rcall recvbyte
    cpi temp,'i'
    breq errorign
    cpi temp,27
    brne errorhalt
    rjmp reset
    
errorign:
    ori flags,(1<<f_ignt1ovf)
    rjmp reset
    
    
font:
.include "font8x8r.inc"
